-- Keyboard routines for ESE by Jeffrey Fielding (JJProg@cyberbury.net)
-- Let me know if you can speed them up. Speed is important.
-- event = AllocateKeyDownEvent(key)
-- event = AllocateKeyUpEvent(key)
-- DeallocateKeyDownEvent(key)
-- DeallocateKeyUpEvent(key)
-- Allocate and deallocate keyboard events. Note that you do not need to
-- keep track of the event. Several handlers may be handling the same
-- key event, so don't remove it. DeallocateKeyUpEvent and
-- DeallocateKeyDownEvent will remove the event when no one is using it.
-- WARNING: If you don't use the AllocateKey...Event functions, your program may
-- not work properly. If you change the variables in this file that track
-- the keys, the event handlers and generators provided here may not work
-- properly. You must call AllocateKey...Event as many times as you call
-- DeallocateKey...Event or it may deallocate the key event before you are
-- done or not deallocate it at all.
include keyread.e
include ese.e
-- Constants for keyboard scan codes
global constant UP_ARROW = 328,
		DOWN_ARROW = 336,
		ENTER = 28,
		RIGHT_ARROW = 333,
		LEFT_ARROW = 331,
		ESC = 1,
		BACKSPACE = 13,
		TAB = 14,
		LEFT_SHIFT = 42,
		RIGHT_SHIFT = 54,
		LEFT_CTRL = 29,
		LEFT_ALT = 56,
		SPACE = 57,
		RIGHT_ALT = 312,
		RIGHT_CTRL = 285,
		F1 = 59,
		F2 = 60,
		F3 = 61,
		F4 = 62,
		F5 = 63,
		F6 = 64,
		F7 = 65,
		F8 = 66,
		F9 = 67,
		F10 = 68,
		F11 = 87,
		F12 = 88
constant NAMES1 = {"1","2","3","4","5","6","7","8","9","0","-","+","BACKSPACE","TAB","q","w","e","r","t","y","u","i","o","p","[","]","ENTER"}
constant NAMES2 = {"a","s","d","f","g","h","j","k","l",";","'"}
constant NAMES3 = {"z","x","c","v","b","n","m",",",".","/","RIGHT SHIFT"}
constant NAMES4 = {"NUM 7","NUM 8","NUM 9","NUM -","NUM 4","NUM 5","NUM 6","NUM +","NUM 1","NUM 2","NUM 3","NUM 0","NUM ."}
-- get the name of a key associated with scan code key
global function keyname(integer key)
    if key >= 59 and key <= 68 then
	return sprintf("F%d",key-58)
    elsif key = 41 then
	return '`'
    elsif key = 42 then
	return "LEFT SHIFT"
    elsif key >= 2 and key <= 28 then
	return NAMES1[key-1]
    elsif key >= 30 and key <= 40 then
	return NAMES2[key-29]
    elsif key >= 44 and key <= 54 then
	return NAMES3[key-43]
    elsif key = 29 then
	return "LEFT CTRL"
    elsif key = 56 then
	return "LEFT ALT"
    elsif key = 57 then
	return "SPACE"
    elsif key = 312 then
	return "RIGHT ALT"
    elsif key = 338 then
	return "INS"
    elsif key = 339 then
	return "DEL"
    elsif key = UP_ARROW then
	return "UP"
    elsif key = DOWN_ARROW then
	return "DOWN"
    elsif key = 333 then
	return "RIGHT"
    elsif key = 331 then
	return "LEFT"
    elsif key = 327 then
	return "HOME"
    elsif key = 329 then
	return "PGUP"
    elsif key = 337 then
	return "PGDN"
    elsif key = 335 then
	return "END"
    elsif key = 43 then
	return "\\"
    elsif key = 87 then
	return "F11"
    elsif key = 88 then
	return "F12"
    elsif key = 285 then
	return "RIGHT CTRL"
    elsif key = 309 then
	return "NUM /"
    elsif key = 55 then
	return "NUM *"
    elsif key >= 71 and key <= 83 then
	return NAMES4[key-70]
    elsif key = 284 then
	return "NUM ENTER"
    elsif key = 1 then
	return "ESC"
    end if
    return ""
end function
-- clear all the keys that have been pressed (or wait until
-- the user stops pressing a key)
global procedure ClearKeyboardBuffer()
    while length(get_keys()) do
    end while
end procedure
-- General key events. Try using the spesific keyboard events with
-- AllocateKeyDownEvent and AllocateKeyUpEvent
global constant KEY_DOWN = AddEvent()
global constant KEY_UP = AddEvent()
sequence lastOn
lastOn = {}
-- Keyboard event generator
procedure KeyCheck()
    sequence keys
    keys = get_keys()
    for i = 1 to length(keys) do
	if not find(keys[i],lastOn) then
	    FireEvent(KEY_DOWN,{keys[i]})
	end if
    end for
    for i = 1 to length(lastOn) do
	if not find(lastOn[i],keys) then
	    FireEvent(KEY_UP,{lastOn[i]})
	end if
    end for
    lastOn = keys
end procedure
constant KEYCHECK = AddEventGenerator(routine_id("KeyCheck"))
sequence hookedKeyDowns, keyDownEvents, keyDownUsers
hookedKeyDowns = {}
keyDownEvents ={}
keyDownUsers = {}
procedure onKeyDown(integer key)
    integer i
    i = find(key,hookedKeyDowns)
    if i then
	FireEvent(keyDownEvents[i],{})
    end if
end procedure
sequence hookedKeyUps, keyUpEvents, keyUpUsers
hookedKeyUps = {}
keyUpEvents = {}
keyUpUsers = {}
procedure onKeyUp(integer key)
    integer i
    i = find(key,hookedKeyUps)
    if i then
	FireEvent(keyUpEvents[i],{})
    end if
end procedure
constant ON_KEY_DOWN = AddEventHandler(KEY_DOWN,routine_id("onKeyDown"))
constant ON_KEY_UP = AddEventHandler(KEY_UP,routine_id("onKeyUp"))
global function AllocateKeyDownEvent(integer key)
    integer i
    i = find(key,hookedKeyDowns)
    if i then
	keyDownUsers[i] += 1
	return keyDownEvents[i]
    else
	i = find(-1,hookedKeyDowns)
	if i then
	    hookedKeyDowns[i] = key
	    keyDownEvents[i] = AddEvent()
	    keyDownUsers[i] = 1
	    return keyDownEvents[i]
	else
	    i = AddEvent()
	    hookedKeyDowns &= key
	    keyDownEvents &= i
	    keyDownUsers &= 1
	    return i
	end if
    end if
end function
global function AllocateKeyUpEvent(integer key)
    integer i
    i = find(key,hookedKeyUps)
    if i then
	keyUpUsers[i] += 1
	return keyUpEvents[i]
    else
	i = find(-1,hookedKeyUps)
	if i then
	    hookedKeyUps[i] = key
	    keyUpEvents[i] = AddEvent()
	    keyUpUsers[i] = 1
	    return keyUpEvents[i]
	else
	    i = AddEvent()
	    hookedKeyUps &= key
	    keyUpEvents &= i
	    keyUpUsers &= 1
	    return i
	end if
    end if
end function
global procedure DeallocateKeyDownEvent(integer key)
    integer i
    i = find(key,hookedKeyDowns)
    if i then
	keyDownUsers[i] -= 1
	if keyDownUsers[i] = 0 then
	    RemoveEvent(keyDownEvents[i])
	    hookedKeyDowns[i] = -1
	end if
    end if
end procedure
global procedure DeallocateKeyUpEvent(integer key)
    integer i
    i = find(key,hookedKeyUps)
    if i then
	keyUpUsers[i] -= 1
	if keyUpUsers[i] = 0 then
	    RemoveEvent(keyUpEvents[i])
	    hookedKeyUps[i] = -1
	end if
    end if
end procedure

