-- Timer code for Gamelib by Jeffrey Fielding (JJProg@cyberbury.net)
-- If anyone has any suggestions for speeding up these routines, please tell me.
-- Speed is very important, so I've tried to speed it up as much as possible.
-- event = AllocateTimerEvent(time)
-- Allocate an event to go off every time seconds. Event is a new ESE event.
-- event = AllocateTimerEventOnce(time)
-- Same as above, but only runs once, then deallocates it.
-- DeallocateTimerEvent(event)
-- Removes the event and disables the timer event generator if it is the last
-- timer event. The generator will be enabled again when AllocateTimerEvent
-- is called.
include ese.e
sequence timerEvents, timerTimes, timerNextTimes, timerOnceOnly
integer allocatedTimerEvents, timerGenerator
timerEvents = {}
timerTimes = {}
timerNextTimes = {}
timerOnceOnly = {}
allocatedTimerEvents = 0
global procedure DeallocateTimerEvent(integer event)
    integer i
    i = find(event,timerEvents)
    if i then
	RemoveEvent(event)
	timerEvents[i] = -1
	allocatedTimerEvents -= 1
	if allocatedTimerEvents = 0 then
	    DisableEventGenerator(timerGenerator)
	end if
    end if
end procedure
procedure TimerGenerator()
    for i = 1 to length(timerEvents) do
	if timerEvents[i] != -1 then
	    if time() >= timerNextTimes[i] then
		FireEvent(timerEvents[i],{})
		if timerOnceOnly[i] then
		    RemoveEvent(timerEvents[i])
		    timerEvents[i] = -1
		    allocatedTimerEvents -= 1
		    if allocatedTimerEvents = 0 then
			RemoveEventGenerator(timerGenerator)
			return
		    end if
		else
		    timerNextTimes[i] = time()+timerTimes[i]
		end if
	    end if
	end if
    end for
end procedure
timerGenerator = AddEventGenerator(routine_id("TimerGenerator"))
DisableEventGenerator(timerGenerator) -- Don't use up extra CPU time
global function AllocateTimerEvent(atom tm)
    integer i, event
    allocatedTimerEvents += 1
    if allocatedTimerEvents = 1 then
	EnableEventGenerator(timerGenerator) -- It had been disabled
    end if
    i = find(-1,timerEvents)
    event = AddEvent()
    if i then
	timerEvents[i] = event
	timerTimes[i] = tm
	timerNextTimes[i] = tm+time()
	timerOnceOnly[i] = 0
	return event
    else
	timerEvents &= event
	timerTimes &= tm
	timerNextTimes &= tm+time()
	timerOnceOnly &= 0
	return event
    end if
end function
global function AllocateTimerEventOnce(atom tm)
    integer i, event
    allocatedTimerEvents += 1
    if allocatedTimerEvents = 1 then
	EnableEventGenerator(timerGenerator) -- It had been disabled
    end if
    i = find(-1,timerEvents)
    event = AddEvent()
    if i then
	timerEvents[i] = event
	timerTimes[i] = tm
	timerNextTimes[i] = tm+time()
	timerOnceOnly[i] = 1
	return event
    else
	timerEvents &= event
	timerTimes &= tm
	timerNextTimes &= tm+time()
	timerOnceOnly &= 1
	return event
    end if
end function

